#ifndef INCLUDE_EXCEPTION_H_DEPRECATED
#define INCLUDE_EXCEPTION_H_DEPRECATED
/*
 * This is the Loris C++ Class Library, implementing analysis, 
 * manipulation, and synthesis of digitized sounds using the Reassigned 
 * Bandwidth-Enhanced Additive Sound Model.
 *
 * Loris is Copyright (c) 1999-2016 by Kelly Fitz and Lippold Haken
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * Exception.h
 *
 * This file formerly defined class Exception, a generic exception class, and 
 * commonly-used derived exception classes, but the name caused build problems
 * on case-insensitive systems having a system header called exception.h, so 
 * the name has been changed to LorisExceptions.h. This file is included as 
 * legacy support. New code should use LorisExceptions.h.
 *
 * Kelly Fitz, 17 Aug 1999
 * loris@cerlsoundgroup.org
 *
 * http://www.cerlsoundgroup.org/Loris/
 *
 */
#include "LorisExceptions.h"

#endif /* ndef INCLUDE_EXCEPTION_H_DEPRECATED */
